@@ -2,27 +2,31 @@
 
 import sys
 import math
-import bisect
+#import bisect
 from collections import defaultdict
 
 import numpy
 from scipy import integrate, interpolate
+from ..integrators.statutils import cumvar, welford, update, finalize
 import itertools
 import functools
 
-try:
-    import healpy
-except:
-    print(" - No healpy - ")
 
-from ..integrators.statutils import cumvar, welford
+import os
+if not( 'RIFT_LOWLATENCY'  in os.environ):
+    # Dont support selected external packages in low latency
+ try:
+    import healpy
+ except:
+    print(" - No healpy - ")
+ try:
+    import vegas
+ except:
+    print(" - No vegas - ")
+
 
 from multiprocessing import Pool
 
-try:
-    import vegas
-except:
-    print(" - No vegas - ")
 
 __author__ = "Chris Pankow <pankow@gravity.phys.uwm.edu>"
 
@@ -150,7 +154,8 @@ class MCSampler(object):
                 self.prior_pdf[params] = lambda x:1
             else:
                 self.prior_pdf[params] = prior_pdf
-        self.prior_pdf[params] = prior_pdf
+        else:
+            self.prior_pdf[params] = prior_pdf
 
         if adaptive_sampling:
             print("   Adapting ", params)
@@ -356,6 +361,9 @@ class MCSampler(object):
         n_adapt -- number of chunks over which to allow the pdf to adapt. Default is zero, which will turn off adaptive sampling regardless of other settings
         convergence_tests - dictionary of function pointers, each accepting self._rvs and self.params as arguments. CURRENTLY ONLY USED FOR REPORTING
 
+        tripwire_fraction - fraction of nmax of iterations after which n_eff needs to be greater than 1+epsilon for epsilon a small number
+        tripwire_epsilon - small number used in tripwire test
+
         Pinning a value: By specifying a kwarg with the same of an existing parameter, it is possible to "pin" it. The sample draws will always be that value, and the sampling prior will use a delta function at that value.
         """
 
@@ -382,7 +390,11 @@ class MCSampler(object):
         # This is a semi-hack to ensure that the integrand is called with
         # the arguments in the right order
         # FIXME: How dangerous is this?
-        args = func.__code__.co_varnames[:func.__code__.co_argcount]
+        if hasattr(func,'__code__'):
+            args = func.__code__.co_varnames[:func.__code__.co_argcount]
+        else:
+            args=list(args[:len(args)])
+            no_protect_params=True   # not code dictionry, so just use arguments in order
 
         #if set(args) & set(params) != set(args):
         # DISABLE THIS CHECK
@@ -405,7 +417,7 @@ class MCSampler(object):
         tempering_exp = kwargs["tempering_exp"] if "tempering_exp" in kwargs else 0.0
         n_adapt = int(kwargs["n_adapt"]*n) if "n_adapt" in kwargs else 0
         floor_integrated_probability = kwargs["floor_level"] if "floor_level" in kwargs else 0
-        temper_log = kwargs["tempering_log"] if "temper_log" in kwargs else False
+        temper_log = kwargs["tempering_log"] if "tempering_log" in kwargs else False
         tempering_adapt = kwargs["tempering_adapt"] if "tempering_adapt" in kwargs else False
         if not tempering_adapt:
             tempering_exp_running=tempering_exp
@@ -427,6 +439,11 @@ class MCSampler(object):
 
         deltalnL = kwargs['igrand_threshold_deltalnL'] if 'igrand_threshold_deltalnL' in kwargs else float("Inf") # default is to return all
         deltaP    = kwargs["igrand_threshold_p"] if 'igrand_threshold_p' in kwargs else 0 # default is to omit 1e-7 of probability
+        bFairdraw  = kwargs["igrand_fairdraw_samples"] if "igrand_fairdraw_samples" in kwargs else False
+        n_extr = kwargs["igrand_fairdraw_samples_max"] if "igrand_fairdraw_samples_max" in kwargs else None
+
+        tripwire_fraction = kwargs["tripwire_fraction"] if "tripwire_fraction" in kwargs else 2  # make it impossible to trigger
+        tripwire_epsilon = kwargs["tripwire_epsilon"] if "tripwire_epsilon" in kwargs else 0.001 # if we are not reasonably far away from unity, fail!
 
         bUseMultiprocessing = kwargs['use_multiprocessing'] if 'use_multiprocessing' in kwargs else False
         nProcesses = kwargs['nprocesses'] if 'nprocesses' in kwargs else 2
@@ -457,6 +474,9 @@ class MCSampler(object):
             bConvergenceTests = False    # if tests are not available, assume not converged. The other criteria will stop it
             last_convergence_test = defaultdict(lambda: False)   # need record of tests to be returned always
         while (eff_samp < neff and self.ntotal < nmax): #  and (not bConvergenceTests):
+            if (self.ntotal > nmax*tripwire_fraction) and (eff_samp < 1+tripwire_epsilon):
+                print(" Tripwire: n_eff too low ")
+                raise Exception("Tripwire on n_eff")
             # Draw our sample points
             args_draw ={}
             if force_no_adapt or save_no_samples:  # don't save permanent sample history if not needed
@@ -544,6 +564,9 @@ class MCSampler(object):
             if bShowEveryEvaluation:
                 for i in range(n):
                     print(" Evaluation details: p,ps, L = ", joint_p_prior[i], joint_p_s[i], fval[i])
+                    print(self.params_ordered)
+                    for indx in numpy.arange(len(args)):
+                        print( self.params_ordered[indx],p_s[indx],p_prior[indx],rv[indx])
 
             # Calculate max L (a useful convergence feature) for debug 
             # reporting.  Not used for integration
@@ -558,8 +581,17 @@ class MCSampler(object):
 
             # running variance
 #            var = cumvar(int_val, mean, var, int(self.ntotal))[-1]
-            var = welford(int_val, mean, var, int(self.ntotal))
-            # running integral
+#            var = welford(int_val, mean, var, int(self.ntotal))
+            if var is None:
+                var=0
+            if mean is None:
+                mean=0
+            current_aggregate = [int(self.ntotal),mean, (self.ntotal-1)*var]
+            current_aggregate = update(current_aggregate, int_val)
+            outvals = finalize(current_aggregate)
+#            print(var, outvals[-1])
+            var = outvals[-1]
+            # running integral (note also in current_aggregate)
             int_val1 += int_val.sum()
             # running number of evaluations
             self.ntotal += n
@@ -622,7 +654,7 @@ class MCSampler(object):
                 if not temper_log:
                     weights = (self._rvs["integrand"][-n_history:]/self._rvs["joint_s_prior"][-n_history:]*self._rvs["joint_prior"][-n_history:])**tempering_exp_running
                 else:
-                    weights = numpy.max([1e-5,numpy.log(self._rvs["integrand"][-n_history:]/self._rvs["joint_s_prior"][-n_history:]*self._rvs["joint_prior"][-n_history:])])**tempering_exp_running
+                    weights = numpy.maximum(1e-5,numpy.log(self._rvs["integrand"][-n_history:] )) #**tempering_exp_running
 
                 if tempering_adapt:
                     # have the adaptive exponent converge to 2/ln(w_max), ln (w)*alpha <= 2. This helps dynamic range
@@ -688,6 +720,7 @@ class MCSampler(object):
             cum_sum = numpy.cumsum(indx_list[:,1])  # find the cumulative sum
             cum_sum = cum_sum/cum_sum[-1]          # normalize the cumulative sum
             indx_list = [int(indx_list[k, 0]) for k, value in enumerate(cum_sum > deltaP) if value]  # find the indices that preserve > 1e-7 of total probability. RECAST TO INTEGER
+
             # FIXME: See previous FIXME
             for key in list(self._rvs.keys()):
                 if isinstance(key, tuple):
@@ -695,6 +728,21 @@ class MCSampler(object):
                 else:
                     self._rvs[key] = self._rvs[key][indx_list]
 
+        # Do a fair draw of points, if option is set
+        if bFairdraw and not(n_extr is None):
+           n_extr = int(numpy.min([n_extr,1.5*eff_samp,1.5*neff]))
+           print(" Fairdraw size : ", n_extr)
+           wt = numpy.array(self._rvs["integrand"]*self._rvs["joint_prior"]/self._rvs["joint_s_prior"]/numpy.max(self._rvs["integrand"]),dtype=float)
+           wt *= 1.0/numpy.sum(wt)
+           if n_extr < len(self._rvs["integrand"]):
+               indx_list = numpy.random.choice(numpy.arange(len(wt)), size=n_extr,replace=True,p=wt) # fair draw
+               # FIXME: See previous FIXME
+               for key in list(self._rvs.keys()):
+                   if isinstance(key, tuple):
+                       self._rvs[key] = self._rvs[key][:,indx_list]
+                   else:
+                       self._rvs[key] = self._rvs[key][indx_list]
+
         # Create extra dictionary to return things
         dict_return ={}
         if convergence_tests is not None:
@@ -717,6 +765,9 @@ def uniform_samp_cdf_inv_vector(a,b,p):
 #uniform_samp_vector = numpy.vectorize(uniform_samp,excluded=['a','b'],otypes=[numpy.float])
 uniform_samp_vector = numpy.vectorize(uniform_samp,otypes=[numpy.float])
 
+def ret_uniform_samp_vector_alt(a,b):
+    return lambda x: numpy.where( (x>a) & (x<b), numpy.reciprocal(b-a),0.0)
+
 # def uniform_samp_withfloor_vector(rmaxQuad,rmaxFlat,pFlat,x):
 #     ret =0.
 #     if x<rmaxQuad:
@@ -741,10 +792,18 @@ def uniform_samp_withfloor_vector(rmaxQuad,rmaxFlat,pFlat,x):
 
 
 # syntatic sugar : predefine the most common distributions
-uniform_samp_phase = numpy.vectorize(lambda x: 1/(2*numpy.pi))
-uniform_samp_psi = numpy.vectorize(lambda x: 1/(numpy.pi))
-uniform_samp_theta = numpy.vectorize(lambda x: numpy.sin(x)/(2))
-uniform_samp_dec = numpy.vectorize(lambda x: numpy.cos(x)/(2))
+uniform_samp_phase = lambda x,numpy=numpy: numpy.broadcast_to(0.5/numpy.pi, numpy.shape(x))
+uniform_samp_psi = lambda x,numpy=numpy: numpy.broadcast_to(1.0/numpy.pi, numpy.shape(x))
+uniform_samp_theta = lambda x,numpy=numpy: 0.5*numpy.sin(x)
+uniform_samp_dec = lambda x,numpy=numpy: 0.5*numpy.cos(x)
+
+uniform_samp_cos_theta = lambda x: numpy.broadcast_to(0.5, numpy.shape(x))
+# uniform_samp_phase = numpy.vectorize(lambda x: 1/(2*numpy.pi))
+# uniform_samp_psi = numpy.vectorize(lambda x: 1/(numpy.pi))
+# uniform_samp_theta = numpy.vectorize(lambda x: numpy.sin(x)/(2))
+# uniform_samp_dec = numpy.vectorize(lambda x: numpy.cos(x)/(2))
+
+# uniform_samp_cos_theta = numpy.vectorize(lambda x: 1./2.)  #dumbest-possible implementation
 
 def quadratic_samp(rmax,x):
         if x<rmax:
@@ -791,8 +850,12 @@ def cos_samp(x):
 def dec_samp(x):
         return numpy.sin(x+numpy.pi/2)/2   # x from 0, pi
 
-cos_samp_vector = numpy.vectorize(cos_samp,otypes=[numpy.float])
-dec_samp_vector = numpy.vectorize(dec_samp,otypes=[numpy.float])
+
+cos_samp_vector = lambda x: cos_samp(numpy.array(x,dtype=numpy.float))
+dec_samp_vector = lambda x: dec_samp(numpy.array(x,dtype=numpy.float))
+
+#cos_samp_vector = numpy.vectorize(cos_samp,otypes=[numpy.float])
+#dec_samp_vector = numpy.vectorize(dec_samp,otypes=[numpy.float])
 def cos_samp_cdf_inv_vector(p):
     return numpy.arccos( 2*p-1)   # returns from 0 to pi
 def dec_samp_cdf_inv_vector(p):
