@@ -30,13 +30,13 @@ import RIFT.lalsimutils as lsu  # problem of relative comprehensive import - dan
 import numpy as np
 try:
   import cupy
-  import optimized_gpu_tools
-  import Q_inner_product
+  from . import optimized_gpu_tools
+  from . import Q_inner_product
   xpy_default=cupy
   junk_to_check_installed = cupy.array(5)  # this will fail if GPU not installed correctly
 except:
-  print(' no cupy')
-  import numpy as cupy
+  print(' no cupy (factored)')
+  cupy=np #import numpy as cupy  # make sure pointer is identical
   optimized_gpu_tools=None
   Q_inner_product=None
   xpy_default=np
@@ -61,30 +61,36 @@ if 'PROFILE' not in os.environ:
 
 __author__ = "Evan Ochsner <evano@gravity.phys.uwm.edu>, R. O'Shaughnessy"
 
-try:
+if not( 'RIFT_LOWLATENCY'  in os.environ):
+  # Dont support external packages in low latency
+ try:
         import NRWaveformCatalogManager3 as nrwf
         useNR =True
         print(" factored_likelihood.py : NRWaveformCatalogManager3 available ")
-except ImportError:
+ except ImportError:
         useNR=False
 
-try:
+ try:
         import RIFT.physics.ROMWaveformManager as romwf
         print(" factored_likelihood.py: ROMWaveformManager as romwf")
         useROM=True
         rom_basis_scale = 1.0*1e-21   # Fundamental problem: Inner products with ROM basis vectors/Sh are tiny. Need to rescale to avoid overflow/underflow and simplify comparisons
-except ImportError:
+ except ImportError:
         useROM=False
         print(" factored_likelihood.py: - no ROM - ")
         rom_basis_scale =1
 
-try:
+ try:
     hasEOB=True
     import RIFT.physics.EOBTidalExternalC as eobwf
-#    import EOBTidalExternal as eobwf
-except:
+ #    import EOBTidalExternal as eobwf
+ except:
     hasEOB=False
     print(" factored_likelihood: no EOB ")
+else:
+  hasEOB=False
+  useROM=False; rom_basis_scale=1
+  useNR=False
 
 distMpcRef = 1000 # a fiducial distance for the template source.
 tWindowExplore = [-0.15, 0.15] # Not used in main code.  Provided for backward compatibility for ROS. Should be consistent with t_ref_wind in ILE.
@@ -216,16 +222,17 @@ def PrecomputeLikelihoodTerms(event_time_geo, t_window, P, data_dict,
     elif (not (NR_group) or not (NR_param)) and  (not use_external_EOB) and (not nr_lookup):
         if not quiet:
                 print( "  FACTORED LIKELIHOOD WITH hlmoff (default ChooseTDModes) " )
-        hlms_list = lsu.hlmoff(P, Lmax) # a linked list of hlms
-        if not isinstance(hlms_list, dict):
-                hlms = lsu.SphHarmFrequencySeries_to_dict(hlms_list, Lmax) # a dictionary
-        else:
-                hlms = hlms_list
-        hlms_conj_list = lsu.conj_hlmoff(P, Lmax)
-        if not isinstance(hlms_list,dict):
-                hlms_conj = lsu.SphHarmFrequencySeries_to_dict(hlms_conj_list, Lmax) # a dictionary
-        else:
-                hlms_conj = hlms_conj_list
+        # hlms_list = lsu.hlmoff(P, Lmax) # a linked list of hlms
+        # if not isinstance(hlms_list, dict):
+        #         hlms = lsu.SphHarmFrequencySeries_to_dict(hlms_list, Lmax) # a dictionary
+        # else:
+        #         hlms = hlms_list
+        # hlms_conj_list = lsu.conj_hlmoff(P, Lmax)
+        # if not isinstance(hlms_list,dict):
+        #         hlms_conj = lsu.SphHarmFrequencySeries_to_dict(hlms_conj_list, Lmax) # a dictionary
+        # else:
+        #         hlms_conj = hlms_conj_list
+        hlms, hlms_conj = lsu.std_and_conj_hlmoff(P,Lmax)
     elif (nr_lookup or NR_group) and useNR:
 	    # look up simulation
 	    # use nrwf to get hlmf
@@ -1686,7 +1693,15 @@ def  DiscreteFactoredLogLikelihoodViaArrayVectorNoLoopOrig(tvals, P_vec, lookupN
 
     return lnL
 
-def  DiscreteFactoredLogLikelihoodViaArrayVectorNoLoop(tvals, P_vec, lookupNKDict, rholmsArrayDict, ctUArrayDict,ctVArrayDict,epochDict,Lmax=2,array_output=False,xpy=np):
+
+# TODO: This should be set based on the marginalization scheme being used.
+# Right now it is hard-coded to not use any tricks, but later will be toggled
+# to use (e.g.) pre-tabulated distance marginaliation dependance.
+def _factored_lnL_helper(kappa_sq, rho_sq):
+    return kappa_sq - 0.5 * rho_sq
+
+
+def  DiscreteFactoredLogLikelihoodViaArrayVectorNoLoop(tvals, P_vec, lookupNKDict, rholmsArrayDict, ctUArrayDict,ctVArrayDict,epochDict,Lmax=2,array_output=False,xpy=np, loglikelihood=_factored_lnL_helper):
     """
     DiscreteFactoredLogLikelihoodViaArray uses the array-ized data structures to compute the log likelihood,
     either as an array vs time *or* marginalized in time. 
@@ -1730,9 +1745,14 @@ def  DiscreteFactoredLogLikelihoodViaArrayVectorNoLoop(tvals, P_vec, lookupNKDic
     deltaT = P_vec.deltaT
 
     # Array to accumulate lnL(t) summed across all detectors.
-    lnL_t_accum = xpy.zeros((npts_extrinsic, npts), dtype=np.float64)
+#    lnL_t_accum = xpy.zeros((npts_extrinsic, npts), dtype=np.float64)
 
-    if xpy is np:
+    # Used to accumulate kappa^2 and rho^2 over all detectors.  They are just
+    # the sum in quadrature of the individual detector contributions.
+    kappa_sq = xpy.zeros((npts_extrinsic, npts), dtype=np.float64)
+    rho_sq = xpy.zeros((npts_extrinsic, npts), dtype=np.float64)
+
+    if (xpy is np) or (optimized_gpu_tools is None):
         simps = integrate.simps
     elif not (xpy is np):
         simps = optimized_gpu_tools.simps
@@ -1798,7 +1818,7 @@ def  DiscreteFactoredLogLikelihoodViaArrayVectorNoLoop(tvals, P_vec, lookupNKDic
         #     Qlms[i] = det_rholms[...,ifirst[i]:ilast[i]].T
 
         # Has shape (npts_extrinsic,)
-        term2 = (
+        rho_sq_det = (
             (F_vec*xpy.conj(F_vec)).real *
             xpy.einsum(
                 "...i,...j,ij",
@@ -1806,14 +1826,18 @@ def  DiscreteFactoredLogLikelihoodViaArrayVectorNoLoop(tvals, P_vec, lookupNKDic
             ).real
         )
 
-        term2 += (
+        rho_sq_det += (
             xpy.square(F_vec) *
             xpy.einsum(
                 "...i,...j,ij",
                 Ylms_vec, Ylms_vec, V,
             )
         ).real
-        term2 *= -0.25 * xpy.square(distMpcRef / distMpc)
+        # NOTE: Double-check this 0.5 is correct.  It used to be -0.25, but
+        # we're slightly reorganizing the expressions from term2 = -0.5 rho_sq
+        # into using rho_sq directly
+        rho_sq_det *= 0.5 * xpy.square(distMpcRef / distMpc)
+
 
         # Has shape (npts_extrinsic, npts).
         # Starts as term1, and accumulates term2 after.
@@ -1857,7 +1881,9 @@ def  DiscreteFactoredLogLikelihoodViaArrayVectorNoLoop(tvals, P_vec, lookupNKDic
             np.conj(FY_dummy_t), Qlms,
             ).real 
 
-        lnL_t_accum += Q_prod_result * (distMpcRef/distMpc)[...,None]
+
+        kappa_sq += Q_prod_result * (distMpcRef/distMpc)[..., np.newaxis]
+        # lnL_t_accum += Q_prod_result * (distMpcRef/distMpc)[...,None]
 
         # lnL_t_accum += Q_inner_product.Q_inner_product_cupy(
         #     FY_conj, Q,
@@ -1867,21 +1893,24 @@ def  DiscreteFactoredLogLikelihoodViaArrayVectorNoLoop(tvals, P_vec, lookupNKDic
 
         # Accumulate term2 into the time-dependent log likelihood.
         # Have to create a view with an extra axis so they broadcast.
-        lnL_t_accum += term2[..., np.newaxis]
+        rho_sq += rho_sq_det[..., np.newaxis]
+        # lnL_t_accum += term2[..., np.newaxis]
 
 #        print lnL_t_accum.shape, lnL_t.shape
 
 #        lnL_t_accum += lnL_t
 
 
+    lnL_t = loglikelihood(kappa_sq, rho_sq)
+
     # Take exponential of the log likelihood in-place.
-    lnLmax  = xpy.max(lnL_t_accum)
-    L_t = xpy.exp(lnL_t_accum - lnLmax, out=lnL_t_accum)
+    lnLmax  = xpy.max(lnL_t)
+    L_t = xpy.exp(lnL_t - lnLmax, out=lnL_t)
 
     L = simps(L_t, dx=deltaT, axis=-1)
 
     # Compute log likelihood in-place.
-    lnL = lnLmax+ xpy.log(L, out=L)
+    lnL = lnLmax + xpy.log(L, out=L)
 
     return lnL
 
@@ -1910,7 +1939,11 @@ def ComputeYlmsArray(lookupNK, theta, phi):
     return Ylms
 
 
-try: 
+fallback=False
+
+if not('RIFT_LOWLATENCY' in os.environ): 
+  # numba is rarely used : we use GPU optimized almost always.  No point incurring these import costs.
+ try:
         import numba
         from numba import vectorize, complex128, float64, int64
         numba_on = True
@@ -1940,7 +1973,10 @@ try:
                         T[indx] = ComputeArrivalTimeAtDetector(det, RA[indx],DEC[indx],  tref)
                 return T
 
-except:
+ except:
+   fallback=True
+
+if fallback or ('RIFT_LOWLATENCY' in os.environ): 
         numba_on = False
         print(" Numba off ")
         # Very inefficient
